/* Sending formatted data packet with nRF24L01. 
 Maximum size of data struct is 32 bytes.
 
 1 - GND
 2 - VCC 3.3V !!! NOT 5V
 3 - CE to Arduino pin 9
 4 - CSN to Arduino pin 10
 5 - SCK to Arduino pin 13
 6 - MOSI to Arduino pin 11
 7 - MISO to Arduino pin 12
 8 - UNUSED
 
 */

#include <SPI.h>
#include <SPI.h>
#include <Wire.h>
#include <nRF24L01.h>
#include <RF24.h>
#include <Servo.h>
#include "printf.h"

struct geoloc {
  float lat;
  float lon;
  geoloc(float x = 0, float y = 0) { lat = x; lon = y; }
};

#ifndef DEGTORAD
#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f
#endif

const uint64_t pipeIn =  0xE8E8F0F0E1LL;

RF24 radio(9, 10);

// The sizeof this struct should not exceed 32 bytes
struct MyData {
  long lon;
  long lat;
  int32_t altitude;
};

MyData data;

Servo yawServo;
Servo pitchServo;

geoloc frameLoc(0,0);
geoloc tagLoc(0,0);

int frameLocInitCounter = 0;
#define FRAMELOC_INIT_SAMPLE_COUNT 30

float distanceToTarget;
float bearingToTarget;
float altitudeToTarget;
boolean newData = false;

// http://www.movable-type.co.uk/scripts/latlong.html
float geoDistance(struct geoloc &a, struct geoloc &b) {
  const float R = 6371000; // km
  float p1 = a.lat * DEGTORAD;
  float p2 = b.lat * DEGTORAD;
  float dp = (b.lat-a.lat) * DEGTORAD;
  float dl = (b.lon-a.lon) * DEGTORAD;

  float x = sin(dp/2) * sin(dp/2) + cos(p1) * cos(p2) * sin(dl/2) * sin(dl/2);
  float y = 2 * atan2(sqrt(x), sqrt(1-x));

  return R * y;
}

float geoBearing(struct geoloc &a, struct geoloc &b) {
  float y = sin(b.lon-a.lon) * cos(b.lat);
  float x = cos(a.lat)*sin(b.lat) - sin(a.lat)*cos(b.lat)*cos(b.lon-a.lon);
  return atan2(y, x) * RADTODEG;
}

void setPitch(float degrees) {
  degrees -= 3; // trim
  degrees = constrain(degrees, -60, 90);
  int servoValue = map(degrees, -90, 90, 135, 45);
  pitchServo.write(servoValue);
}

void setYaw(float degrees) {
  degrees *= 0.9;
  degrees += -6; // trim
  degrees = constrain(degrees, -170, 170);
  int servoValue = map(degrees, -180, 180, 135, 45);
  yawServo.write(servoValue);
}

/**************************************************/

void setup() {
  Serial.begin(9600);
  printf_begin();
  yawServo.attach(3);
  pitchServo.attach(5);
  setPitch(0);
  setYaw(0);
  
  radio.begin();
  radio.setDataRate(RF24_250KBPS);
  radio.setAutoAck(false);
  radio.openReadingPipe(1,pipeIn);
  radio.startListening();
  
  radio.printDetails();
   
}


/**************************************************/

void recvData()
{  
  while ( radio.available() ) {
    radio.read(&data, sizeof(MyData));
    tagLoc.lat = data.lat / 10000000.0f;
    tagLoc.lon = data.lon / 10000000.0f;
    
    newData = true;
    
    if ( frameLocInitCounter < FRAMELOC_INIT_SAMPLE_COUNT ) {
      float uptake = 1 - frameLocInitCounter / (float)FRAMELOC_INIT_SAMPLE_COUNT;
      frameLoc.lat = uptake * tagLoc.lat + (1-uptake) * frameLoc.lat;
      frameLoc.lon = uptake * tagLoc.lon + (1-uptake) * frameLoc.lon;
      frameLocInitCounter++;
      Serial.println(data.lat);   
      Serial.println(frameLoc.lat);      
    }
  }
}

/**************************************************/

void loop() {
  
  float frameHeading = 0;
  
  recvData();
  
  if ( newData ) {
    distanceToTarget = geoDistance( frameLoc, tagLoc );
    bearingToTarget = geoBearing( frameLoc, tagLoc );
    altitudeToTarget = data.altitude / 100.0f;
  }
  
  float localBearing = bearingToTarget - frameHeading;
  while (localBearing < -180) localBearing += 360;
  while (localBearing >  180) localBearing -= 360;
  
  setYaw(localBearing);


Serial.print(distanceToTarget);
Serial.print(" ");
Serial.println(localBearing);  
  


  float pitch = 0;
  if ( distanceToTarget != 0 )
    pitch = atan( altitudeToTarget / distanceToTarget );
  setPitch(pitch * RADTODEG);
  
}








